
{$B-,R-,S-,V-,F-,I-,A+}
{$G+,R-,S-,M 1024, 0, 0 }
program irman;


uses crt,dos;

const  coms       = 4;     { Anzahl der untersttzten Schnittstellen }
       NUMIRCODES    = 30;

       ua         : array[1..coms] of word = ($3f8,$2f8,$3e8,$2e8);
       datainout  = 0;     { UART-Register-Offsets }
       intenable  = 1;
       intids     = 2;     { Read  }
       fifoctrl   = 2;     { Write }
       linectrl   = 3;
       modemctrl  = 4;
       linestat   = 5;
       modemstat  = 6;
       scratch    = 7;

       UartNone   = 0;     { Ergebnisse von ComType }
       Uart8250   = 1;
       Uart16450  = 2;
       Uart16550  = 3;
       Uart16550A = 4;

       NoFifo     = $00;   { Triggerlevel bei 16550-Chips }
       FifoTL1    = $07;
       FifoTL4    = $47;
       FifoTL8    = $87;
       FifoTL14   = $C7;

type   paritype   = (Pnone,Podd,Pxxxx,Peven);   { mgliche Paritts-Typen }
       code_array = record
         key:byte;
         key2:byte;
         rc:longint;
       end;

{ Parameter fr Schnittstelle einstellen
{ no       : Nummer  (1-4)
  address  : I/O-Adresse, 0 -> Adresse wird beibehalten
  _irq     : Interrupt-Nummer  (z.B. 3 fr IRQ3, 4 fr IRQ4); 0..15 }


const  active     : array[1..coms] of boolean = (false,false,false,false);
       irq        : array[1..coms] of byte = ($04,$03,0,0);
       intmask    : array[1..coms] of byte = ($10,$08,0,0);
       intcom2    : array[1..coms] of boolean = (false,false,false,false);

       MS_CTS     = $10;       { Modem-Status-Register }
       MS_DSR     = $20;
       MS_RI      = $40;       { Ring Indicator: Klingelsignal }
       MS_DCD     = $80;       { Data Carrier Detect           }
       MC_DTR     = $01;       { Modem Control Register }
       MC_RTS     = $02;

type   bufft      = array[0..65534] of byte;

var    savecom    : array[1..coms] of pointer;
{       exitsave   : pointer;}
       bufsize    : array[1..coms] of word;
       buffer     : array[1..coms] of ^bufft;
       bufi,bufo  : array[1..coms] of word;
       irfound    : boolean;
       Ircodes    : array[1..NumIRCodes] of code_array;
       setup:boolean;
       i:integer;
       numcodes:byte;
       comport:byte;


procedure error(text:string);
begin
  writeln('UART Error: ',text);
end;

function strs(l:longint):string;
var s : string;
begin
  str(l,s);
  strs:=s;
end;


{--- Interrupt-Handler -----------------------------------------------}


procedure sendbyte(no,b:byte);              { Byte senden }
begin
  while (port[ua[no]+linestat] and $20) = 0 do;
  port[ua[no]]:=b;
end;

procedure flushinput(no:byte);            { Receive-Puffer lschen }
begin
  bufo[no]:=bufi[no];
end;

procedure DropDtr(no:byte);                 { DTR=0 setzen      }
begin
  port[ua[no]+modemctrl]:=port[ua[no]+modemctrl] and (not MC_DTR);
end;

procedure SetDtr(no:byte);                  { DTR=1 setzen      }
begin
  port[ua[no]+modemctrl]:=port[ua[no]+modemctrl] or MC_DTR;
end;

procedure DropRts(no:byte);                 { RTS=0 setzen      }
begin
  port[ua[no]+modemctrl]:=port[ua[no]+modemctrl] and (not MC_RTS);
end;

procedure SetRts(no:byte);                  { RTS=1 setzen      }
begin
  port[ua[no]+modemctrl]:=port[ua[no]+modemctrl] or MC_RTS;
end;

procedure cli; inline($fa);            { Interrupts sperren   }
procedure sti; inline($fb);            { Interrupts freigeben }

function received(no:byte):boolean;      { Testen, ob Daten vorhanden }
begin
  received:=(bufi[no]<>bufo[no]);
end;


function receive(no:byte; var b:byte):boolean;   { Byte holen, falls vorh. }
begin
  if bufi[no]=bufo[no] then
    receive:=false
  else begin
    b:=buffer[no]^[bufo[no]];
    inc(bufo[no]);
    if bufo[no]=bufsize[no] then bufo[no]:=0;
    receive:=true;
    end;
end;

function GetCode:word;
var i:byte;
    w:word;
    c:byte;
  begin
    w:= 0;
    for i := 1 to 6 do begin
     repeat until received(comport);
      if receive(comport,c) then begin
        if c <> 0 then begin
          w := w shl 8;
          w := w + c;
        end;
      end;
    end;
      delay(5);
      flushinput(comport);
{    writeln(w);}
    getcode := w;
  end;

function readkey2:integer;
 var c:char;
     b:byte;
  begin
    asm
      mov ah,0
      int $16
      mov c,al
      mov b,ah
    end;
    readkey2:= b*256 + ord(c);
  end;

procedure IRLoadConfig(fn:string);
  var
    t:text;
    s:string;
    s1,s2:string;
    key:byte;
    i:integer;
    next:integer;
    code:word;
    code1,code2:byte;
    ch:char;


  begin
    if (fn <> '') then
    begin
      next := 0;
      {$I-}
      assign(t,fn);
      reset(t); {$I+}
      if IOResult = 0 then
      begin
        while not eof(t) do
        begin
          readln(t,s);
          s1 := copy(s,1,1);
          if s1 <> ';' then begin { Comment }
            s2 := copy(s,1,pos('=',s)-1);
            inc(next);
            s1 := copy(s2,1,pos(',',s2)-1);
            s2 := copy(s2,pos(',',s2)+1,length(s2));
            val(s2,key,i);
            val(s1,code,i);
            if code <> 0 then
            begin
              ircodes[next].key2 := key;
              ircodes[next].key := code;
            end else
            begin
              ircodes[next].key2 := code;
              ircodes[next].key := key;
            end;
            s2 := copy(s,pos('=',s)+1,length(s));
            val(s2,code,i);
            ircodes[next].rc := code;
          end;
        end;
        close(t);
      end;
      numcodes := next;
    end;
  end;

procedure IRSaveConfig(fn:string);
  var t:text;
  begin
    if fn <> '' then
    begin
     {$I-}
      assign(t,fn);
      rewrite(t);  {$I+}
      if IOResult = 0 then
        begin
          for i := 1 to numcodes do
            writeln(t,ircodes[i].key,',',ircodes[i].key2,'=',ircodes[i].rc);
           close(t);
        end;
    end;
  end;

procedure IRSetup(com:byte);
var t:text;
    s:string;
    s1,s2:string;
    key:byte;
    ke:integer;
    i:integer;
    next:integer;
    code:longint;
    numkeys:integer;
    code1,code2:byte;
    ch:char;


    procedure addkey2(key1,key2:byte;pos:integer);
      begin
        bufi[com] := 0;bufo[com] := 0;
        writeln('Push button for key');
        repeat until bufi[com]=6;
        code := getcode;
        bufi[com]:=0;
        bufo[com]:=0;
        ircodes[numcodes].key := key1;
        ircodes[numcodes].key2 := key2;
        ircodes[numcodes].rc := code;
      end;

      function AddAnotherKey:boolean;
         var c:char;
        begin
          writeln(numcodes, ' are currently assigned.');
          write('Add Another Key? [Y/N] ');
          repeat
            repeat until keypressed;
              c:=upcase(readkey);
          until c in ['Y','N'];
          writeln(c);
          if c = 'Y' then AddAnotherKey := true else AddAnotherKey := false;
        end;
begin
  for i := 1 to numircodes do begin
    IRCODES[i].key := 0;
    IRCODES[i].key2 := 255;
    IRCODES[i].rc := 0;
  end;
  next := 0;

  assign(t,'irsetup.cfg');
  {$I-} reset(t); {$I+}
  if IOResult <> 0 then begin
    writeln('Cannot find config file!');
    writeln;
    writeln('We will now create a configuration file.');
    writeln('This process allows you to associate keys with buttons on the remote');
    writeln;
    numcodes := 0;
    setup := true;
    while (AddAnotherKey) do begin
      inc(numcodes);
      writeln('Select the key on the keyboard.');

      repeat until keypressed;
      ke := readkey2;
      code2 := (ke and $ff);
      code1 := (ke and $ff00) div 256;
      addkey2(code1,code2,numkeys);
    end;
    IRSaveConfig('irsetup.cfg');
  end else close(t);
  IRLoadConfig('irsetup.cfg');
  writeln(numcodes,' codes loaded.');
end;


procedure initIR(comport:byte);
var ci,cr :byte;
  begin
    dropdtr(comport);
   droprts(comport);
   delay(100);
   setdtr(comport);
   setrts(comport);
   delay(100);
   flushinput(comport);
   delay(50);
   sendbyte(comport,ord('I'));
   delay(1);
   sendbyte(comport,ord('R'));
   delay(10);
   if received(comport) then begin
     receive(comport,ci);
     receive(comport,cr);
     if (chr(ci) = 'O') and (chr(cr) = 'K') then begin
       writeln('IR found!');
       IRSetup(comport);
       irfound := true;
     end;
   end;
   delay(5);
   flushinput(comport);
   bufi[comport]:=0;
   bufo[comport]:=0;
 end;

procedure DoIR(com:byte);
var li:longint;
    count:integer;

procedure stuffKey(c : char; scan : byte); assembler;
asm
  mov cl, c
  mov ch, scan
  mov ah, 5
  int $16
  {al=0 if success, 1 if failed, but we just ignore that here}
end;

procedure stuffChar(c : char); assembler;
asm
  mov cl,c
  xor ch,ch
  mov ah,5
  int $16
end;

   begin
     if received (com) then begin
       li := getcode;
       count := 1;
         while (ircodes[count].rc <> 0) do begin
           if (li = ircodes[count].rc) then
                 stuffkey(chr(ircodes[count].key2),ircodes[count].key);
           inc(count);
         end;
       bufi[com]:=0;
       bufo[com]:=0;
     end;

   end;

procedure com1server; interrupt;
begin
  if intcom2[1] then port[$a0]:=$20;
  port[$20]:=$20;                      { Interrupt-Controller resetten }
  buffer[1]^[bufi[1]]:=port[ua[1]];
  inc(bufi[1]); if bufi[1]=bufsize[1] then bufi[1]:=0;
  if (irfound) and (bufi[1]=6) then  DoIR(1);
end;

procedure com2server; interrupt;
begin
  if intcom2[2] then port[$a0]:=$20;
  port[$20]:=$20;
  buffer[2]^[bufi[2]]:=port[ua[2]];
  inc(bufi[2]); if bufi[2]=bufsize[2] then bufi[2]:=0;
  if (irfound) and (bufi[2]=6) then  DoIR(2);
end;

procedure com3server; interrupt;
begin
  if intcom2[3] then port[$a0]:=$20;
  port[$20]:=$20;
  buffer[3]^[bufi[3]]:=port[ua[3]];
  inc(bufi[3]); if bufi[3]=bufsize[3] then bufi[3]:=0;
  if (irfound) and (bufi[3]=6) then  DoIR(3);
end;

procedure com4server; interrupt;
begin
  if intcom2[4] then port[$a0]:=$20;
  port[$20]:=$20;
  buffer[4]^[bufi[4]]:=port[ua[4]];
  inc(bufi[4]); if bufi[4]=bufsize[4] then bufi[4]:=0;
  if (irfound) and (bufi[4]=6) then  DoIR(4);
end;

procedure com1FIFOserver; interrupt;
begin
  if port[ua[1]+intids] and 4<>0 then
    repeat
      buffer[1]^[bufi[1]]:=port[ua[1]];
      inc(bufi[1]); if bufi[1]=bufsize[1] then bufi[1]:=0;
    until not odd(port[ua[1]+linestat]);
  if intcom2[1] then port[$a0]:=$20;
  port[$20]:=$20;                      { Interrupt-Controller resetten }
  if (irfound) and (bufi[1]=6) then  DoIR(1);
end;

procedure com2FIFOserver; interrupt;
begin
  if port[ua[2]+intids] and 4<>0 then
    repeat
      buffer[2]^[bufi[2]]:=port[ua[2]];
      inc(bufi[2]); if bufi[2]=bufsize[2] then bufi[2]:=0;
    until not odd(port[ua[2]+linestat]);
  if intcom2[2] then port[$a0]:=$20;
  port[$20]:=$20;
  if (irfound) and (bufi[2]=6) then  DoIR(2);
end;

procedure com3FIFOserver; interrupt;
begin
  if port[ua[3]+intids] and 4<>0 then
    repeat
      buffer[3]^[bufi[3]]:=port[ua[3]];
      inc(bufi[3]); if bufi[3]=bufsize[3] then bufi[3]:=0;
    until not odd(port[ua[3]+linestat]);
  if intcom2[3] then port[$a0]:=$20;
  port[$20]:=$20;
  if (irfound) and (bufi[3]=6) then  DoIR(3);
end;

procedure com4FIFOserver; interrupt;
begin
  if port[ua[4]+intids] and 4<>0 then
    repeat
      buffer[4]^[bufi[4]]:=port[ua[4]];
      inc(bufi[4]); if bufi[4]=bufsize[4] then bufi[4]:=0;
    until not odd(port[ua[4]+linestat]);
  if intcom2[4] then port[$a0]:=$20;
  port[$20]:=$20;
  if (irfound) and (bufi[4]=6) then  DoIR(4);
end;


function ComType(no:byte):byte;     { Typ des UART-Chips ermitteln }
var uart        : word;
    lsave,ssave : byte;
    isave,iir   : byte;
begin
  uart:=ua[no];
  lsave:=port[uart+linectrl];
  port[uart+linectrl]:=lsave xor $ff;
  if port[uart+linectrl]<>lsave xor $ff then
    ComType:=UartNone
  else begin
    port[uart+linectrl]:=lsave;
    ssave:=port[uart+scratch];
    port[uart+scratch]:=$5a;
    if port[uart+scratch]<>$5a then
      ComType:=Uart8250                 { kein Scratchpad vorhanden }
    else begin
      port[uart+scratch]:=$a5;
      if port[uart+scratch]<>$a5 then
        ComType:=Uart8250               { kein Scratchpad vorhanden }
      else begin
        isave:=port[uart+intids];
        port[uart+fifoctrl]:=1;
        iir:=port[uart+intids];
        if isave and $80=0 then port[uart+fifoctrl]:=0;
        if iir and $40<>0 then ComType:=Uart16550A
        else if iir and $80<>0 then ComType:=Uart16550
        else ComType:=Uart16450;
        end;
      end;
    port[uart+scratch]:=ssave;
    end;
end;


{--- Schnitte einstellen / aktivieren / freigeben --------------------}

procedure SetComParams(no:byte; address:word; _irq:byte);
begin
  if (no>=1) and (no<=coms) then begin
    if address<>0 then ua[no]:=address;
    irq[no]:=_irq;
    intmask[no]:=(1 shl (_irq and 7));
    intcom2[no]:=(_irq>7);      { 2. Interrupt-Controller }
    end;
end;

procedure setuart(comno:byte; baudrate:longint; parity:paritype;
                  wlength,stops:byte);
var uart : word;
begin
  uart:=ua[comno];
  port[uart+linectrl]:=$80;
  port[uart+datainout]:=lo(word(115200 div baudrate));
  port[uart+datainout+1]:=hi(word(115200 div baudrate));
  port[uart+linectrl]:= (wlength-5) or (stops-1)*4 or ord(parity)*8;
  port[uart+modemctrl]:=$0b;
  if port[uart+datainout]<>0 then;      { dummy }
end;


procedure clearstatus(no:byte);
begin
  if port[ua[no]+datainout]<>0 then;               { dummy-Read }
  if port[ua[no]+linestat]<>0 then;
  if port[ua[no]+modemstat]<>0 then;
  if intcom2[no] then port[$a0]:=$20;
  port[$20]:=$20;
end;


function IntNr(no:byte):byte;
begin
  if irq[no]<8 then IntNr:=irq[no]+8
  else IntNr:=irq[no]+$68;
end;

procedure ActivateCom(no:byte; buffersize:word; FifoTL:Byte);
var p : pointer;
    i : byte;
begin
  if active[no] then begin
    error('Error activating comport '+strs(no)+'!');
    exit;
    end
  else if (no<1) or (no>coms) or (irq[no]=0) then
    error('Comport '+strs(no)+' is not a valid comport!')
  else
    active[no]:=true;

  bufsize[no]:=buffersize;                 { Puffer anlegen }
  getmem(buffer[no],buffersize);
  bufi[no]:=0; bufo[no]:=0;
  fillchar(buffer[no]^,bufsize[no],0);

  IF (fifotl > 0)
    THEN BEGIN
           Port[(ua[no] + fifoctrl)] := fifotl;
           IF ((Port[(ua[no] + intids)] AND $40) = 0)
             THEN BEGIN
                    Port[(ua[no] + fifoctrl)] := 0;
                    fifotl := NoFifo;
                  END;
         END;

  IF (fifotl > 0)
    THEN CASE no OF
           1 : p:=@com1FIFOserver;
           2 : p:=@com2FIFOserver;
           3 : p:=@com3FIFOserver;
           4 : p:=@com4FIFOserver;
         END {CASE}
    ELSE CASE no OF
           1 : p:=@com1server;
           2 : p:=@com2server;
           3 : p:=@com3server;
           4 : p:=@com4server;
         END; {CASE}

  getintvec(IntNr(no),savecom[no]);           { IRQ setzen }
  setintvec(IntNr(no),p);
  port[ua[no]+intenable]:=$01;                     { Int. bei Empfang }
  if intcom2[no] then
    port[$a1]:=port[$a1] and (not intmask[no])     { Ints freigeben }
  else
    port[$21]:=port[$21] and (not intmask[no]);
  clearstatus(no);
  initir(no);
  if IRFound then
  begin
    writeln('IR TSR installed.');
    SwapVectors;

    keep(0);

  end;
end;


procedure releasecom(no:byte);
begin
  if not active[no] then
    error('Error releaseing comport '+strs(no)+'!')
  else begin
    active[no]:=false;
    port[ua[no]+intenable]:=0;
    if intcom2[no] then
      port[$a1]:=port[$a1] or intmask[no]    { Controller: COMn-Ints sperren }
    else
      port[$21]:=port[$21] or intmask[no];
    port[ua[no]+fifoctrl]:=0;
    setintvec(IntNr(no),savecom[no]);
    clearstatus(no);
    freemem(buffer[no],bufsize[no]);
    end;
end;


{ Exit-Prozedur }



{--- Daten senden / empfangen ----------------------------------------}


function peek(no:byte; var b:byte):boolean;
begin
  if bufi[no]=bufo[no] then
    peek:=false
  else begin
    b:=buffer[no]^[bufo[no]];
    peek:=true;
    end;
end;

procedure hsendbyte(no,b:byte);           { Byte senden, mit CTS-Handshake }
begin
  while (port[ua[no]+modemstat] and $10) = 0 do;
  while (port[ua[no]+linestat] and $20) = 0 do;
  port[ua[no]]:=b;
end;

procedure putbyte(no,b:byte);             { Byte im Puffer hinterlegen }
begin
  if bufo[no]=0 then bufo[no]:=bufsize[no]
  else dec(bufo[no]);
  buffer[no]^[bufo[no]]:=b;
end;



{--- Modem-Status-Lines ----------------------------------------------}

function rring(no:byte):boolean;            { Telefon klingelt  }
begin
  rring:=(port[ua[no]+modemstat] and MS_RI)<>0;
end;

function carrier(no:byte):boolean;          { Carrier vorhanden }
begin
  carrier:=(port[ua[no]+modemstat] and MS_DCD)<>0;
end;



{ True -> Modem (oder entsprechendes Gert)  ist bereit, Daten zu empfangen }

function GetCTS(no:byte):boolean;
begin
  getcts:=((port[ua[no]+modemstat] and $10)<>0) and
           ((port[ua[no]+linestat] and $20)<>0);
end;


function ticker:longint;
begin
  ticker:=meml[Seg0040:$6c];
end;

procedure SendBreak(no:byte);             { Break-Signal      }
var teiler : word;
    btime  : longint;
    t0     : longint;
begin
  CLI;
  port[ua[no]+linectrl]:=port[ua[no]+linectrl] or $80;
  teiler:=port[ua[no]] + 256*port[ua[no]+1];
  port[ua[no]+linectrl]:=port[ua[no]+linectrl] and $7f;
  STI;
  btime:=teiler DIV 200;
  IF (btime<1) THEN btime:=1;
  t0:=ticker;
  inc(btime,ticker);
  Port[ua[no]+linectrl]:=port[ua[no]+linectrl] or $40;   { set break }
  repeat
  until (ticker>btime) or (ticker<t0);
  Port[ua[no]+linectrl]:=port[ua[no]+linectrl] and $bf;  { clear break }
end;


  var p:paritype;
begin
  if paramcount <> 1 then begin
    writeln('Usage: IRMAN comport');
    writeln('Example: IRMAN 1');
    writeln;
    halt(1);
  end;
  val(paramstr(1),comport,i);
  irfound := false;
  setup := false;
  p := Pnone;
  writeln('Searching for IR device on comport ',comport);
  setuart(comport,9600,p,8,1);
  activatecom(comport,20,0);
end.
